package pl.edu.icm.yadda.remoting.cli;

import java.security.InvalidParameterException;

import org.apache.commons.cli.CommandLine;
import org.opensaml.lite.xacml.ctx.DecisionType;
import org.springframework.context.ConfigurableApplicationContext;

import pl.edu.icm.yadda.aas.client.ISecuritySession;
import pl.edu.icm.yadda.aas.client.LoginResult;
import pl.edu.icm.yadda.aas.client.SecuritySessionImpl;
import pl.edu.icm.yadda.aas.client.authn.IAuthenticationManager;
import pl.edu.icm.yadda.service2.aas.AAError;
import pl.edu.icm.yadda.service2.user.token.AnonymousToken;
import pl.edu.icm.yadda.service2.user.token.LoginPasswordToken;
import pl.edu.icm.yadda.service2.usersession.ISessionService;

/**
 * Context helper class.
 * @author mhorst
 *
 */
public abstract class CommonContextHelper {

	public static String BEAN_AUTHN_MANAGER = "authnManager";
	public static String BEAN_SESSION_SERVICE = "AasSessionService";
	public static String BEAN_IMPORTER_FACTORY = "importerFactory";
	public static String PROCESS_MANAGER_FACADE = "processManagerFacade";
	public static String PROCESS_MANAGER_FACADE3 = "processManagerFacade3";
	public static String BEAN_DOMAIN = "dl-domain";
	
	/**
	 * Provides domain by checking among commandline attributes of extracting from
	 * repository descriptor when available. Returns null when domain not specified in either place.
	 * @param cli
	 * @param context
	 * @return
	 */
	public static String getDomain(CommandLine cli, ConfigurableApplicationContext context) {
		if (cli.hasOption(CliParameters.CLI_SHORT_PARAM_DOMAIN)) {
			return cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_DOMAIN);
		} else if (context.containsBean(BEAN_DOMAIN)) {
			String domain = (String) context.getBean(BEAN_DOMAIN);
			System.out.println("using descriptor domain: " + domain);
			return domain;
		} else {
			return null;
		}
	}
	
	public static String buildSessionId(String opType, String userId, String domain) {
		StringBuilder strBuilder = new StringBuilder();
		strBuilder.append(opType);
		strBuilder.append("-");
		if (userId!=null) {
			strBuilder.append(userId);
		} else {
			strBuilder.append("unauthenticatedUser");
		}
		if (domain!=null) {
			strBuilder.append('@');
			strBuilder.append(domain);
		}
		strBuilder.append("_");
		strBuilder.append(System.currentTimeMillis());
		return strBuilder.toString();
	}
	
	
	/**
	 * Picks suitable context path.
	 * @param cli
	 * @return suitable context path
	 */
	public static String pickContextPath(CommandLine cli) {
		if (cli.hasOption(CliParameters.CLI_SHORT_PARAM_CTX_FILE)) {
			return cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_CTX_FILE);
		} else {
//			checking whether credentials provided
			if (cli.hasOption(CliParameters.CLI_SHORT_PARAM_USER) || 
					cli.hasOption(CliParameters.CLI_SHORT_PARAM_IPADDR)) {
				return "classpath:pl/edu/icm/yadda/remoting/cli/cli-context-secure-discovered.xml";	
			} else {
				return "classpath:pl/edu/icm/yadda/remoting/cli/cli-context-discovered.xml";
			}
		}
	}
	
	/**
	 * Handles security session by either logging in user when 
	 * details provided or initializing anonymous session.
	 * @param sessionIdPrefix 
	 * @param cli
	 * @param context
	 * @param allowAnonymousIpAuthn
	 * @return true when session maintained successfully, false otherwise
	 */
	public static boolean maintainSession(String sessionIdPrefix, 
			CommandLine cli, ConfigurableApplicationContext context,
			boolean allowAnonymousIpAuthn) {
//    	authn block
    	if (cli.hasOption(CliParameters.CLI_SHORT_PARAM_USER)) {
    		if (cli.hasOption(CliParameters.CLI_SHORT_PARAM_PASSWD)) {
    			String domain = getDomain(cli, context);
//	    		connecting and initializing session
    			ISessionService sessionService = (ISessionService) context.getBean(
    					CommonContextHelper.BEAN_SESSION_SERVICE);
    			sessionService.connect(CommonContextHelper.buildSessionId(sessionIdPrefix,
    					cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_USER),
    					domain));
    			ISecuritySession session = new SecuritySessionImpl(
    					sessionService.getSessionId(), null, cli.hasOption(CliParameters.CLI_SHORT_PARAM_IPADDR)?
    							cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_IPADDR):null);
    			sessionService.bind(
    					ISecuritySession.SECURITY_SESSION_ATTRIBUTE_NAME, session);
    			
//	    		authenticating when login/password pair provided
    			IAuthenticationManager authnManager = (IAuthenticationManager)context.getBean(
    					CommonContextHelper.BEAN_AUTHN_MANAGER);
    			LoginResult result = authnManager.login(new LoginPasswordToken(
    					cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_USER), 
    					cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_PASSWD),
    					domain));
    			if (result.getDecition() != DecisionType.DECISION.Permit) {
    				System.err.println("Authentication for user " + cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_USER) + 
    						" failed, decision is " + result.getDecition());
    				if (!result.getErrors().isEmpty()) {
    					System.err.println("The following errors occurred:");
	    				for (AAError error : result.getErrors()) {
	    					System.err.println(error.toString());
	    					if (error.getThrowable() != null) {
	    						System.err.println("Exception caught: " + error.getThrowable());
	    					}
	    				}	
    				}
    				return false;
    			} else {
    				return true;
    			}
    		} else {
    			throw new InvalidParameterException(
    					"no password provided for user " + cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_USER));
    		}
		} else if (allowAnonymousIpAuthn && cli.hasOption(CliParameters.CLI_SHORT_PARAM_IPADDR)) {
//			connecting and initializing session
			ISessionService sessionService = (ISessionService) context.getBean(
					CommonContextHelper.BEAN_SESSION_SERVICE);
			sessionService.connect(CommonContextHelper.buildSessionId(sessionIdPrefix,
					null,
					cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_IPADDR)));
			ISecuritySession session = new SecuritySessionImpl(
					sessionService.getSessionId(), null, cli.getOptionValue(
							CliParameters.CLI_SHORT_PARAM_IPADDR));
			sessionService.bind(
					ISecuritySession.SECURITY_SESSION_ATTRIBUTE_NAME, session);
			
//			authenticating when login/password pair provided
			IAuthenticationManager authnManager = (IAuthenticationManager)context.getBean(
					CommonContextHelper.BEAN_AUTHN_MANAGER);
			LoginResult result = authnManager.login(new AnonymousToken(
					cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_IPADDR)));
			if (result.getDecition() != DecisionType.DECISION.Permit) {
				System.err.println("Authentication for ip " + cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_IPADDR) + 
						" failed, decision is " + result.getDecition());
				if (!result.getErrors().isEmpty()) {
					System.err.println("The following errors occurred:");
    				for (AAError error : result.getErrors()) {
    					System.err.println(error.toString());
    					if (error.getThrowable() != null) {
    						System.err.println("Exception caught: " + error.getThrowable());
    					}
    				}	
				}
				return false;
			} else {
               System.err.println("LoginResult: decision=" + result.getDecition()
				+ ", errors.size()=" + result.getErrors().size()+", assertion issuer: " + result.getAssertion().getIssuer().toString());
               return true;
            }
		} else {
//			initializing session when no authentication performed
			ISessionService sessionService = (ISessionService) context.getBean(
					CommonContextHelper.BEAN_SESSION_SERVICE);
			sessionService.connect(CommonContextHelper.buildSessionId(sessionIdPrefix, null, null));
			ISecuritySession session = new SecuritySessionImpl(
					sessionService.getSessionId(), null, null);
			sessionService.bind(
					ISecuritySession.SECURITY_SESSION_ATTRIBUTE_NAME, session);
			return true;
		}
	}
}
