package pl.edu.icm.yadda.remoting.cli;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.springframework.beans.factory.BeanFactoryUtils;
import org.springframework.context.ConfigurableApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;

import pl.edu.icm.yadda.process.IProcessListener;
import pl.edu.icm.yadda.process.ProcessContext;
import pl.edu.icm.yadda.process.ProcessingStats;

/**
 * A CLI tool for running local processes (defined in the CLI, not in the repository).
 *
 * @author mpol
 */
public final class ProcessRunner {

	/**
	 * @param args first argument is the process name to run, the rest are process dependent
	 */
	public static void main(String[] args) {
		String syntax = "ProcessRunner [-h|<process-name> <process-param>...]";
		HelpFormatter help = new HelpFormatter();
		help.setOptionComparator(new Comparator<Option>() {
			public int compare(Option o1, Option o2) {
				return o1.getLongOpt().compareTo(o2.getLongOpt());
			}
		});

		Options opts = new Options();
		opts.addOption("h", "help", false, "print this help message");

		@SuppressWarnings("static-access")
		Option paramsOpt = OptionBuilder.withDescription("add a key=value pair to the process context")
			.hasArgs(2).withArgName("key=value").withValueSeparator()
			.withLongOpt("context-param").create('P');
		opts.addOption(paramsOpt);

		@SuppressWarnings("static-access")
		Option flagsOpt = OptionBuilder.withDescription("add a key=true pair to the process context")
			.hasArg().withArgName("key")
			.withLongOpt("context-flag").create('F');
		opts.addOption(flagsOpt);

		try {
			CommandLine cli = new GnuParser().parse(opts, args);
			if (cli.hasOption('h')) {
				help.printHelp(syntax, opts);
				System.exit(0);
			}

			ConfigurableApplicationContext ctx = new ClassPathXmlApplicationContext("classpath:processes/*.xml");
			ctx.registerShutdownHook();

			String[] argsRest = cli.getArgs();

			if (argsRest.length == 0) {
				System.out.println("Available processes:");
				for (String name : BeanFactoryUtils.beanNamesForTypeIncludingAncestors(ctx, IProcessFacade.class))
					System.out.println(name);
				System.exit(0);
			}

			@SuppressWarnings("unchecked")
			Map<String, IProcessFacade> procs = BeanFactoryUtils.beansOfTypeIncludingAncestors(ctx, IProcessFacade.class);

			ProcessRunner pr = new ProcessRunner(procs);

			pr.run(argsRest[0],
				Arrays.asList(Arrays.copyOfRange(argsRest, 1, argsRest.length)),
				ctxMap(cli.getOptionProperties(paramsOpt.getOpt()), cli.getOptionProperties(flagsOpt.getOpt())));

		} catch (Exception e) {
			System.err.println("Error: "+e.getMessage());
			System.exit(1);
		}
	}

	/**
	 * Converts multiple <code>Properties</code> objects to a <code>Map&lt;String,Serializable></code>.
	 *
	 * @param propss multiple <code>Properties</code> objects
	 * @return a map containing the properties
	 */
	private static HashMap<String, Serializable> ctxMap(Properties... propss) {
		HashMap<String, Serializable> ctxMap = new HashMap<String, Serializable>();
		for (Properties props : propss)
			for (String key : props.stringPropertyNames())
				ctxMap.put(key, props.getProperty(key));
		return ctxMap;
	}

	private Map<String, IProcessFacade> procs;

	/**
	 * Creates a process runner with a collection of available processes.
	 *
	 * @param procs known processes
	 */
	public ProcessRunner(Map<String, IProcessFacade> procs) {
		this.procs = procs;
	}

	/**
	 * Creates and runs a process.
	 *
	 * @param name name of the process to run
	 * @param params parameters for the process creator
	 * @param ctxMap data to put in the process context map
	 * @return processing stats from the finished process
	 */
	private ProcessingStats run(String name, List<String> params, HashMap<String, Serializable> ctxMap) {
		IProcessFacade proc = procs.get(name);
		//XXX: to be generalized if ProcessRunner is ever used elsewhere.
		IProcessListener listener = newListener();
		return proc.run(new ProcessContext(listener, ctxMap), params, System.in);
	}

	/**
	 * A static factory of process listeners writing progress data to the standard error stream.
	 *
	 * @return a process listener
	 */
	private static IProcessListener newListener() {
		return new IProcessListener() {

			double input;
			double output;
			double percent;

			public void sourceInput(double volume) {
				input += volume;
			}

			public void sinkOutput(double volume) {
				output += volume;
				double newp = 100*output/input;
				if (newp-percent > 10) {
					percent = newp;
					System.err.format("%d%% of input processed", (int) Math.floor(percent));
				}
			}

			public void propertyChanged(String property, Serializable value) {
				System.err.format("Process property '%s' changed to '%s'", property, value);
			}

			public void processStarted() {
				System.err.println("Process started");
			}

			public void processFinished(boolean cancelled) {
				System.err.format("Process %s%n", cancelled?"cancelled":"finished");
			}

			public void notifyEvent(String operation, String[] args, String resultCode,	String result) {
				System.err.format("Process event: operation=%s, args=%s, resultCode=%s, result=%s%n",
					operation, Arrays.toString(args), resultCode, result);
			}
		};
	}
}
