/**
 * @author tzr
 *
 */
package pl.edu.icm.yadda.remoting.cli;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Serializable;
import java.net.URL;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.prefs.Preferences;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.joda.time.format.ISODateTimeFormat;
import org.springframework.context.ConfigurableApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;

import pl.edu.icm.yadda.process.IProcessListener;
import pl.edu.icm.yadda.process.ProcessingStats;
import pl.edu.icm.yadda.process.harvester.Constants;
import pl.edu.icm.yadda.service2.process.IProcessManager;
import pl.edu.icm.yadda.service2.process.MultiplexingProcessListener;
import pl.edu.icm.yadda.service2.process.Process;
import pl.edu.icm.yadda.service2.process.Processor;
import pl.edu.icm.yadda.service2.process.progress.ProcessProgress;
import pl.edu.icm.yadda.service2.process.progress.ProcessProgressMonitor;
import pl.edu.icm.yadda.service2.process.progress.TaskProgress;

public class RunProcess {
       static class InvalidOptionException extends Exception {

		/**
		 * @param msg the error message
		 */
		public InvalidOptionException(String msg) {
			super(msg);
		}

		/**
		 * Creates an exception without any error message
		 */
		public InvalidOptionException() {}		
	}
	private static final String CLASSPATH_PREFIX	= "classpath:";
	private static final String PARAM_DATE_FORMAT = "compatible with ISO8601, e.g. yyyy-MM-dd'T'HH:mm:ss.SSSZZ";
	
        
	/**
	 * Lists processes currently running on the specified repository
	 * Args:
	 * <ul>
	 *  <li>[repositoryUrl]</li>
	 * </ul>
	 * @param args
	 */
	public static void main(String[] args) throws Exception {

		Options options = new Options();
		options.addOption("from", true, "start date parameter of the process ("
				+ PARAM_DATE_FORMAT + ")");
		options.addOption("to", true, "end date of parameter the process ("
				+ PARAM_DATE_FORMAT + ")");
		options.addOption("cutoff", true,
				"cutoff date of parameter the process (" + PARAM_DATE_FORMAT
						+ ")");
		options.addOption(CliParameters.CLI_SHORT_PARAM_SERV_DESCR_LOC, true, "repository service descriptor url");
		options.addOption("p", true, "process to be run");		
		options.addOption(CliParameters.CLI_SHORT_PARAM_HELP, false, "show help");
		options.addOption(CliParameters.CLI_SHORT_PARAM_CTX_FILE, true, "specify custom context file");
		
        options.addOption(CliParameters.CLI_SHORT_PARAM_USER, "login", true, "user login to be used when importing to secured respository");
		options.addOption(CliParameters.CLI_SHORT_PARAM_PASSWD, "password", true, "user password to be used when importing to secured respository");
		options.addOption(CliParameters.CLI_SHORT_PARAM_DOMAIN, "domain", true, "user domain, optional");
		options.addOption(CliParameters.CLI_SHORT_PARAM_IPADDR, "ip", true, "ip");
		
                Option property  = OptionBuilder.withArgName( "key=value" ).hasArgs().
			withDescription( "provide entry in the process context. Multiple entries could be provided." ).create( "P" );
		options.addOption(property);
		
		Option propertyWithFileLocAsValue  = OptionBuilder.withArgName( "key=value" ).hasArgs().
		withDescription( "provide file content entry in the process context. " +
				"The 'value' of key/value pair is read from file, which location is stored in 'value'. " +
				"Multiple entries could be provided." ).create( "C" );
		options.addOption(propertyWithFileLocAsValue);
				
		CommandLineParser parser = new GnuParser();
		
		CommandLine cli = parser.parse(options, args);

		if (!cli.getArgList().isEmpty()) {
		    System.err.println("Unexpected parameters: "+cli.getArgList());
		    System.exit(1);
		}
		
		if (!cli.hasOption(CliParameters.CLI_SHORT_PARAM_SERV_DESCR_LOC) || !cli.hasOption("p") || cli.hasOption(CliParameters.CLI_SHORT_PARAM_HELP)) {
			HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("run.sh", options);
			return;
		}
											
		String prid = cli.getOptionValue("p");
		
		String[] parameters = cli.getOptionValues("P");
//		support for reading file content as parameter value
		String[] paramsWithFileContents = cli.getOptionValues("C");

		Date d1, d2, cutoff;
		try {
			d1 = cli.hasOption("from") ? ISODateTimeFormat
					.dateOptionalTimeParser()
					.parseDateTime(cli.getOptionValue("from"))
					.toCalendar(Locale.getDefault()).getTime()
					: new GregorianCalendar(1980, 1, 1).getTime();
		} catch (IllegalArgumentException e) {
			System.out.println("Invalid processing start date "
					+ cli.getOptionValue("from"));
			return;
		}
		try {
			d2 = cli.hasOption("to") ? ISODateTimeFormat
					.dateOptionalTimeParser()
				.parseDateTime(cli.getOptionValue("to"))
				.toCalendar(Locale.getDefault()).getTime()
				: new GregorianCalendar(3000, 1, 1).getTime();
		} catch (IllegalArgumentException e) {
			System.out.println("Invalid processing end date "
					+ cli.getOptionValue("to"));
			return;
		}
		try {
			cutoff = cli.hasOption("cutoff") ? ISODateTimeFormat
					.dateOptionalTimeParser()
					.parseDateTime(cli.getOptionValue("cutoff"))
					.toCalendar(Locale.getDefault()).getTime() : null;
		} catch (IllegalArgumentException e) {
			System.out.println("Invalid cutoff date "
					+ cli.getOptionValue("cutoff"));
			return;
		}
		
		String  descriptorLocation= cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_SERV_DESCR_LOC);				
		
		//Preferences.userRoot().put("serviceRootUrl", serviceRootUrl);				
		String contextPath = CommonContextHelper.pickContextPath(cli);
		System.out.println("using context path: " + contextPath);
		
		Preferences.userRoot().put("descriptorLocation", descriptorLocation);
	        ConfigurableApplicationContext context = 
	        	new ClassPathXmlApplicationContext(contextPath);			
       // ConfigurableApplicationContext context = 
        //	new ClassPathXmlApplicationContext("classpath:pl/edu/icm/yadda/remoting/cli/cli-context.xml");
                
	    Map<String, Serializable> contextMap = new HashMap<String, Serializable>();

        try {
        	if (!CommonContextHelper.maintainSession("process", cli, context, true)) {
        		System.exit(1);
        	}
//	    	execute only when session successfully handled
//			this is facility feature to use same login/pass pair within AuthorizeNode 
//			when no explicit parameters provided (will be eventually overwritten by 
//			forthcoming params processing)
	    	if (cli.hasOption(CliParameters.CLI_SHORT_PARAM_USER)) {
	    		contextMap.put(Constants.PARAM_AUTH_LOGIN, cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_USER));
				contextMap.put(Constants.PARAM_AUTH_PASSWORD, cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_PASSWD));
				contextMap.put(Constants.PARAM_AUTH_DOMAIN, cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_DOMAIN));
	    	}
			if (cli.hasOption(CliParameters.CLI_SHORT_PARAM_IPADDR)) {
				contextMap.put(Constants.PARAM_AUTH_IP, cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_IPADDR));
			}
	    	
//	    	process block  
	    	IProcessManager pm = (IProcessManager)context.getBean(
	    			CommonContextHelper.PROCESS_MANAGER_FACADE);
	        Processor<?> processor = pm.processor(prid);
	        
	        if (processor == null){
	        	System.out.println("Processor " + prid + " not found");
	        	return;
	        }
	        
	        contextMap.put(ProcessProgressMonitor.CONTEXT_KEY,
	        		new ProcessProgressMonitor());
	    	
	    	contextMap.put(Constants.PARAM_FROM, d1);
	    	contextMap.put(Constants.PARAM_TO, d2);
	    	
	    	if (parameters != null) {
		    	for (String parameter : parameters) {
		    		String key;
		    		String value;
		    		if (parameter.contains("=")) {
		    			key = parameter.substring(0, parameter.indexOf("="));
		    			value = parameter.substring(parameter.indexOf("=") + 1);
		    		} else {
		    			key = parameter;
		    			value = "true";
		    		}
		    		contextMap.put(key, value);
		    	}
	    	}
//			support for reading file content as parameter value
	    	if (paramsWithFileContents != null) {
		    	for (String parameter : paramsWithFileContents) {
		    		String key;
		    		String value;
		    		if (parameter.contains("=")) {
		    			key = parameter.substring(0, parameter.indexOf("="));
		    			value = readFileContent(
		    					parameter.substring(parameter.indexOf("=") + 1));
		    		} else {
		    			key = parameter;
		    			value = "true";
		    		}
		    		contextMap.put(key, value);
		    	}
	    	}
	    	
	    	if (cutoff != null) contextMap.put(Constants.PARAM_FORCE_CUTOFF_DATE, cutoff);
	        
	        Process process = processor.withContextMap(contextMap).submit();
	        
	        IProcessListener l = process.getListener();
	        if (l instanceof MultiplexingProcessListener) {
	            System.out.println("Hurray, good listener!!!!!");
	            ((MultiplexingProcessListener) l).addProcessListener(new MyProcessListener(prid));
	        } else {
	            System.out.println("Buuuuu - bad listener....");
	        }
	        
	        ProcessingStats stats = process.get();
	        
	        printStats(prid, stats);
        
        } finally {
        	context.close();
        }
	}
	
	protected static void printStats(String process, ProcessingStats stats) {
		System.out.println("Process " + process);
		for (Object stat : stats.values()) {
			System.out.println(stat);
		}
		System.out.println();
		
		for (String targetName : stats.getTargets()) {
			printStats(targetName, stats.getTargetStats(targetName));
		}
	}
	
	/**
	 * Reads file content from file of given location.
	 * Supports reading classpath files.
	 * @param fileLoc
	 * @return file content
	 */
	protected static String readFileContent(String fileLoc) {
		if (fileLoc.startsWith(CLASSPATH_PREFIX)) {
			fileLoc = fileLoc.substring(CLASSPATH_PREFIX.length());
		}
		File file = null;
//		added support for reading file content by web application
		URL fileURL = RunProcess.class.getResource("/"+fileLoc);
		if (fileURL!=null) {
			file = new File(fileURL.getPath());
		} else {
			file = new File(fileLoc);
		}
		FileInputStream fis = null;
		try {
			fis = new FileInputStream(file);
			BufferedReader reader = new BufferedReader(new InputStreamReader(fis));
			StringBuffer strBuff = new StringBuffer();
			String thisLine = null;
			while ((thisLine = reader.readLine()) != null) {  
				strBuff.append(thisLine);
			}
			return strBuff.toString();
		} catch (IOException e) {
			throw new RuntimeException(e);
		} finally {
			if (fis!=null) {
				try {
					fis.close();
				} catch (IOException e) {
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	private static class MyProcessListener
    implements IProcessListener {
		
		protected String taskName;
		
		public MyProcessListener(String taskName) {
			this.taskName = taskName;
		}
	
		public void processStarted() {
		}
		
		public void processFinished(boolean cancelled) {
		}
		
		public void sourceInput(double volume) {
		}
		
		public void sinkOutput(double volume) {
		}
		
		public void propertyChanged(String property, Serializable value) {
		    if (ProcessProgressMonitor.DEFAULT_PROCESS_PROGRESS_PROPERTY_NAME.equals(property)) {
		        ProcessProgress p = (ProcessProgress) value;		        
		        for (TaskProgress tp : p.getProgress().values()) {
		        	System.out.println("Current progress of task " + tp.getTaskId() + " is " + tp.getCurrent() + " of " + tp.getTotal());	
		        }
		        
		    }
		}

		public void notifyEvent(String operation, String[] args,
				String resultCode, String result) {
			// TODO Auto-generated method stub
			
		}
	}

}
