package pl.edu.icm.yadda.remoting.cli.service2.user;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.security.InvalidParameterException;
import java.util.ArrayList;
import java.util.Formatter;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.prefs.Preferences;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;
import org.springframework.context.ConfigurableApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;
import org.springframework.util.StringUtils;

import pl.edu.icm.yadda.aas.userdb.DTOListAdapter;
import pl.edu.icm.yadda.aas.userdb.UserDBInitializer;
import pl.edu.icm.yadda.aas.userdb.UserDTO;
import pl.edu.icm.yadda.common.pagination.PaginationResult;
import pl.edu.icm.yadda.remoting.cli.CliParameters;
import pl.edu.icm.yadda.remoting.cli.CommonContextHelper;
import pl.edu.icm.yadda.service2.user.UserCatalog;
import pl.edu.icm.yadda.service2.user.UserCatalogConstants;
import pl.edu.icm.yadda.service2.user.UserCatalogFacade;
import pl.edu.icm.yadda.service2.user.model.Group;
import pl.edu.icm.yadda.service2.user.model.UserAttributesConstants;
import pl.edu.icm.yadda.service2.user.model.UserData;

/**
 * Users manager command line utility.
 * @author mhorst
 *
 */
public class ManageUsers {
	
	public static final String SCRIPT_NAME = "manage_users.sh";
	
	public static final String PROPERTY_AAS_PROVIDER_URL = "aas.provider.url";
	
	public static final String BEAN_USER_CATALOG_FACADE = "usersCatalogFacade";
	
	public static final String DEFAULT_ADMIN_NAME = "admin";
	
	public static final String ATTR_KEY_CREATED_BY = "created-by";

	public static final String ACTION_INSERT = "insert";
	public static final String ACTION_REMOVE = "remove";
	public static final String ACTION_LIST_CREATED = "list-created";
	public static final String ACTION_LIST_ALL = "list-all";
	public static final String ACTION_PURGE = "purge";
	
	public static final String CLI_SHORT_PARAM_ACTION = "a";
	public static final String CLI_SHORT_PARAM_FILE = "f";
	public static final String CLI_SHORT_PARAM_OVERWRITE = "o";
	public static final String CLI_SHORT_PARAM_TARGET_DOMAIN = "td";
	public static final String CLI_SHORT_PARAM_TARGET_USER = "tu";

	/**
	 * Lists processes currently running on the specified repository
	 * @param args
	 */
	public static void main(String[] args) throws Exception {

		Options options = new Options();
		options.addOption(CliParameters.CLI_SHORT_PARAM_HELP, false, "show help");
		options.addOption(CliParameters.CLI_SHORT_PARAM_SERV_DESCR_LOC, true, "specify shared AAS and userdb root context location, REQUIRED!");
		options.addOption(CliParameters.CLI_SHORT_PARAM_USER, "login", true, "admin login to be used when authenticating to AAS, REQUIRED!");
		options.addOption(CliParameters.CLI_SHORT_PARAM_PASSWD, "password", true, "admin password to be used when authenticating to AAS, REQUIRED!");
		options.addOption(CliParameters.CLI_SHORT_PARAM_DOMAIN, "domain", true, "admin domain, optional, set to '" + UserCatalogConstants.SUPER_DOMAIN_NAME + "' by default");
//		manager specific
		options.addOption(CLI_SHORT_PARAM_ACTION, "action", true, "action to be performed, REQUIRED! One of: " +
				"'" + ACTION_INSERT + "','" + ACTION_REMOVE + "','" + ACTION_PURGE + "','"+ ACTION_LIST_ALL + "','" + ACTION_LIST_CREATED + "'");
		options.addOption(CLI_SHORT_PARAM_FILE, "file", true, "csv file location containing list of users to be created, required for action '" + ACTION_INSERT + "'");
		options.addOption(CLI_SHORT_PARAM_OVERWRITE, "overwrite", false, "flag indicating existing entries should be overwriten, optional for action '" + ACTION_INSERT + "'");
		options.addOption(CLI_SHORT_PARAM_TARGET_DOMAIN, "target-domain", true, "fully qualified target domain, required for all actions excluding '" + ACTION_INSERT + "'");
		options.addOption(CLI_SHORT_PARAM_TARGET_USER, "target-user", true, "user name to be removed, required for action '" + ACTION_REMOVE + "'");
		
		CommandLineParser parser = new PosixParser();
		CommandLine cli = null;
		ConfigurableApplicationContext context = null;
		try {			
			cli = parser.parse(options, args);
			
			if (!cli.getArgList().isEmpty()) {
			    System.err.println("Unexpected parameters: " + cli.getArgList());
			    System.exit(1);
			}
			
			String contextPath = "classpath:pl/edu/icm/yadda/remoting/cli/cli-context-secure-aas-userdb-bundle.xml";
			System.out.println("using context path: " + contextPath);
			
			if (!cli.hasOption(CLI_SHORT_PARAM_ACTION) ||
					!cli.hasOption(CliParameters.CLI_SHORT_PARAM_SERV_DESCR_LOC) || 
					cli.hasOption(CliParameters.CLI_SHORT_PARAM_HELP)) {
				HelpFormatter formatter = new HelpFormatter();
				formatter.printHelp(SCRIPT_NAME, options);
				return;
			}
			
			String rootContextLocation = cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_SERV_DESCR_LOC);
			Preferences.userRoot().put(PROPERTY_AAS_PROVIDER_URL, rootContextLocation);
	        context = new ClassPathXmlApplicationContext(contextPath);
	        
//			maintains session
        	System.out.println("logging in admin user: '" + 
        			cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_USER) + "'");
	        if (!CommonContextHelper.maintainSession("manage-users", cli, context, true)) {
        		System.exit(1);
        	}
	        
//	    	user-manage block
	        final UserCatalog userCatalogFacade = (UserCatalogFacade) context.getBean(
	        		BEAN_USER_CATALOG_FACADE);
	        
	        if (ACTION_INSERT.equals(cli.getOptionValue(CLI_SHORT_PARAM_ACTION))) {
	        	if (cli.hasOption(CLI_SHORT_PARAM_FILE)) {
	        		String fileName = cli.getOptionValue(CLI_SHORT_PARAM_FILE);
	        		BufferedReader in = null;
	        		List<UserData> userDataList = new ArrayList<UserData>(); 
	                try {
	        			in = new BufferedReader(new InputStreamReader(new FileInputStream(
	        					new File(fileName))));
	                    String line;
	                    while ((line = in.readLine()) != null) {
//	                    	csv order: domain;user-name;passwd;email;role1,role2;attr1,attr2
	                    	String trimmed = line.trim();
	                    	if (trimmed.length()==0 || trimmed.startsWith("#")) {
	                    		continue;
	                    	}
	                    	String[] csvArray = StringUtils.delimitedListToStringArray(trimmed, ";");
	                    	if (csvArray.length==6) {
	                    		UserDTO userDTO = new UserDTO();
		                    	userDTO.setDomain(csvArray[0]);
		                    	userDTO.setName(csvArray[1]);
		                    	userDTO.setPassword(csvArray[2]);
		                    	userDTO.setEmail(csvArray[3]);
		                    	userDTO.setRolesCSV(csvArray[4]);
		                    	userDTO.setAttrsCSV(csvArray[5]);
//		                    	inserting creator attribute
		                    	UserData userData = DTOListAdapter.convertUser(
		                    			userDTO, true);
		                    	if (userData.getUser().getAttributes()==null) {
		                    		userData.getUser().setAttributes(new HashMap<String, String>());
		                    	}
		            			userData.getUser().getAttributes().put(
		            					ATTR_KEY_CREATED_BY, generateCreatedByAttrValue(cli, context));
		                    	userDataList.add(userData);
		                    	
	                    	} else {
	                    		System.err.println("invalid csv line: '" + line + 
	                    				"', expected 6 tokens delimited with ';'!");
	                    		System.exit(1);
	                    	}
	                    }
//	                    storing objects
	                    System.out.println("inserting " + userDataList.size() + " users...");
	                    UserDBInitializer initializer = new UserDBInitializer();
	                    initializer.setInitActive(true);
	                    initializer.setUserCatalog(userCatalogFacade);
	                    initializer.setOverwriteExistingObjects(cli.hasOption(CLI_SHORT_PARAM_OVERWRITE));
	                    initializer.setUsers(userDataList);
	                    initializer.afterPropertiesSet();
	                    System.out.println("all users stored");
	                } finally {
	                	try {
	                    	if (in!=null)
	            				in.close();
            			} catch (IOException e) {
            				System.err.println("Couldn't close "+fileName+" file!");
            				e.printStackTrace(System.err);
            			}
	                }

	        	} else {
	        		System.err.println("required parameter '" + CLI_SHORT_PARAM_FILE + "' is missing!");
		        	HelpFormatter formatter = new HelpFormatter();
					formatter.printHelp(SCRIPT_NAME, options);
					System.exit(1);
	        	}
	        } else if (ACTION_REMOVE.equals(cli.getOptionValue(CLI_SHORT_PARAM_ACTION))) {
	        	if (cli.hasOption(CLI_SHORT_PARAM_TARGET_DOMAIN) && cli.hasOption(CLI_SHORT_PARAM_TARGET_USER)) {
	        		userCatalogFacade.deleteUser(
	        				cli.getOptionValue(CLI_SHORT_PARAM_TARGET_USER), 
	        				cli.getOptionValue(CLI_SHORT_PARAM_TARGET_DOMAIN));
	        		System.out.println("user '" + cli.getOptionValue(CLI_SHORT_PARAM_TARGET_USER) + 
	        				"' successfully removed from domain '"+ cli.getOptionValue(CLI_SHORT_PARAM_TARGET_DOMAIN) +"'");
	        		return;
	        	} else {
	        		System.err.println("both parameters '" + CLI_SHORT_PARAM_TARGET_DOMAIN + "' and '" + 
	        				CLI_SHORT_PARAM_TARGET_USER + "' are required!");
		        	HelpFormatter formatter = new HelpFormatter();
					formatter.printHelp(SCRIPT_NAME, options);
					System.exit(1);
	        	}
	        } else if (ACTION_PURGE.equals(cli.getOptionValue(CLI_SHORT_PARAM_ACTION))) {
	        	if (cli.hasOption(CLI_SHORT_PARAM_TARGET_DOMAIN)) {
	        		String domain = cli.getOptionValue(CLI_SHORT_PARAM_TARGET_DOMAIN);
//	        		removing users
	        		PaginationResult<UserData> users = userCatalogFacade.searchUsers(
	        				domain, null, null, null, null, 0, -1);
	        		if (users!=null && users.getResults()!=null && 
	        				users.getResults().size()>0) {
	        			System.out.println("removing " + users.getTotalCount() + 
		        				" users from domain " + domain);
	        			for (UserData user : users.getResults()) {
	        				userCatalogFacade.deleteUser(user.getId(), domain);
	        			}
	        		} else {
	        			System.out.println("no users to remove from domain " + domain);
	        		}
//	        		removing groups
	        		PaginationResult<Group> groups = userCatalogFacade.listGroups(domain, 0, -1);
	        		if (groups!=null && groups.getResults()!=null &&
	        				groups.getResults().size()>0) {
	        			System.out.println("removing " + groups.getTotalCount() + 
		        				" groups from domain " + domain);
		        		for (Group group : groups.getResults()) {
//		        			no cascading to prevent GroupNotFound exceptions
		        			userCatalogFacade.deleteGroup(group.getGroupName(), false);
		        		}	
	        		} else {
	        			System.out.println("no groups to remove from domain " + domain);
	        		}
	        		
	        	} else {
	        		System.err.println("parameter '" + CLI_SHORT_PARAM_TARGET_DOMAIN + "' is required!");
		        	HelpFormatter formatter = new HelpFormatter();
					formatter.printHelp(SCRIPT_NAME, options);
					System.exit(1);
	        	}
	        	
	        	
	        	
	        } else if (ACTION_LIST_ALL.equals(cli.getOptionValue(CLI_SHORT_PARAM_ACTION)) ||
	        		ACTION_LIST_CREATED.equals(cli.getOptionValue(CLI_SHORT_PARAM_ACTION))) {
	        	if (cli.hasOption(CLI_SHORT_PARAM_TARGET_DOMAIN)) {
	        		PaginationResult<UserData> results = userCatalogFacade.searchUsers(
	        				cli.getOptionValue(CLI_SHORT_PARAM_TARGET_DOMAIN), 
	        				null, null, prepareSearchAttributes(cli, context), null, 0, -1);
	        		int idx = 0;
	        		System.out.println(new Formatter().format("%1$-3s %2$-15s %3$-15s %4$-25s %5$-60s %6$s",
    	        			"nr", "domain", "userid", "email", "roles", "attributes"));
	        		for (UserData user : results.getResults()) {
	        			System.out.println(new Formatter().format("%1$-3s %2$-15s %3$-15s %4$-25s %5$-60s %6$s",
	    	        			++idx, 
	    	        			user.getUser().getDomain(), 
	    	        			user.getUser().getIdentifiers().iterator().next(),
	    	        			user.getUser().getAttributes().get(UserAttributesConstants.EMAIL),
	    	        			user.getUser().getRoles(),
	    	        			prepareAttributes(user.getUser().getAttributes(),
	    	        					UserAttributesConstants.EMAIL)));
	        		}
	        		
	        	} else {
	        		System.err.println("required parameter '" + CLI_SHORT_PARAM_TARGET_DOMAIN + "' is missing!");
		        	HelpFormatter formatter = new HelpFormatter();
					formatter.printHelp(SCRIPT_NAME, options);
					System.exit(1);
	        	}
	        } else {
	        	System.err.println("invalid '" + CLI_SHORT_PARAM_ACTION + "' parameter: " +
	        			cli.getOptionValue(CLI_SHORT_PARAM_ACTION));
	        	HelpFormatter formatter = new HelpFormatter();
				formatter.printHelp(SCRIPT_NAME, options);
				System.exit(1);
	        }
			
		} catch (ParseException e) {
			throw new InvalidParameterException(e.getMessage());
		} finally {
			if (context!=null) {
				context.close();
			}
		}
	}
	
	protected static Map<String, String> prepareAttributes(Map<String, String> source, 
			String... attrKeysToBeSkipped) {
		if (source!=null && !source.isEmpty()) {
			Map<String, String> result = new HashMap<String, String>(source);
			for (String currentAttr : attrKeysToBeSkipped) {
				result.remove(currentAttr);
			}
			return result;
		} else {
			return source;	
		}
	}
	
	static String generateCreatedByAttrValue(CommandLine cli, ConfigurableApplicationContext context) {
		StringBuffer strBuff = new StringBuffer();
		strBuff.append(CommonContextHelper.getDomain(cli, context));
		strBuff.append(':');
		strBuff.append(cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_USER));
		return strBuff.toString();
	}
	
	static Map<String, String> prepareSearchAttributes(CommandLine cli, ConfigurableApplicationContext context) {
		if (ACTION_LIST_CREATED.equals(cli.getOptionValue(CLI_SHORT_PARAM_ACTION))) {
			Map<String, String> result = new HashMap<String, String>();
			result.put(ATTR_KEY_CREATED_BY, generateCreatedByAttrValue(cli, context));
			return result;
		} else {
			return null;
		}
	}
}
