package pl.edu.icm.yadda.remoting.cli.service3.process;

import java.security.InvalidParameterException;
import java.util.Date;
import java.util.Formatter;
import java.util.Map.Entry;
import java.util.Set;
import java.util.prefs.Preferences;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;
import org.springframework.context.ConfigurableApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;

import pl.edu.icm.yadda.remoting.cli.CliParameters;
import pl.edu.icm.yadda.remoting.cli.CommonContextHelper;
import pl.edu.icm.yadda.service2.exception.ServiceException;
import pl.edu.icm.yadda.service3.process.ProcessFacade;
import pl.edu.icm.yadda.service3.process.stats.ErrorAwareProcessStats;
import pl.edu.icm.yadda.service3.process.stats.ErrorEntry;

/**
 * Shows process statistics.
 * @author mhorst
 *
 */
public class ProcessStats {

	public static final String CLI_RUN_PROCESS_SHORT_PARAM_ERRS = "e";

	/**
	 * Lists process details working on Yadda-repository.
	 * @param args
	 */
	public static void main(final String[] args) throws Exception {
		final Options options = new Options();
		options.addOption(CliParameters.CLI_SHORT_PARAM_SERV_DESCR_LOC, true, "specify repository descriptor location, required");
		options.addOption(CliParameters.CLI_SHORT_PARAM_CTX_FILE, true, "specify custom context file, optional");
		options.addOption(CliParameters.CLI_SHORT_PARAM_PID, true, "identifier of the process, required");
		options.addOption(CliParameters.CLI_SHORT_PARAM_USER, "login", true, "user login to be used when authenticating to secured respository");
		options.addOption(CliParameters.CLI_SHORT_PARAM_PASSWD, "password", true, "user password to be used when authenticating to secured respository");
		options.addOption(CliParameters.CLI_SHORT_PARAM_DOMAIN, "domain", true, "user domain, optional");
		options.addOption(CliParameters.CLI_SHORT_PARAM_HELP, false, "show help");
		options.addOption(CLI_RUN_PROCESS_SHORT_PARAM_ERRS, false, "print errors details");

		final CommandLineParser parser = new PosixParser();
		CommandLine cli = null;
		ConfigurableApplicationContext context = null;
		String pId = null;
		try {
			cli = parser.parse(options, args);

			if (!cli.getArgList().isEmpty()) {
			    System.err.println("Unexpected parameters: " + cli.getArgList());
			    System.exit(1);
			}

			final String contextPath = CommonContextHelper.pickContextPath(cli);
			System.out.println("using context path: " + contextPath);

			if (!cli.hasOption(CliParameters.CLI_SHORT_PARAM_SERV_DESCR_LOC) ||
					!cli.hasOption(CliParameters.CLI_SHORT_PARAM_PID) ||
					cli.hasOption(CliParameters.CLI_SHORT_PARAM_HELP)) {
				final HelpFormatter formatter = new HelpFormatter();
				formatter.printHelp("psstat.sh", options);
				return;
			}

			final String descriptorLocation = cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_SERV_DESCR_LOC);
			Preferences.userRoot().put("descriptorLocation", descriptorLocation);
			pId = cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_PID);
			context = new ClassPathXmlApplicationContext(contextPath);

//			maintains session
			if (!CommonContextHelper.maintainSession("process-stats", cli, context, false)) {
        		System.exit(1);
        	}

//	    	process block
			try {
				final ProcessFacade facade = (ProcessFacade) context.getBean(
						CommonContextHelper.PROCESS_MANAGER_FACADE3);
				final ErrorAwareProcessStats stats = facade.getErrorAwareProcessStats(pId);
				if (stats!=null) {
					System.out.println(new Formatter().format("%1$-20s %2$-30s",
							"start time", new Date(stats.getStartTime())));
					System.out.println(new Formatter().format("%1$-20s %2$-20s",
							"status", stats.getStatus()));
					final int[] procCount = stats.getProcessedCount();
					System.out.println(new Formatter().format("%1$-20s %2$-20s",
							"processed count", procCount[0]));
					System.out.println(new Formatter().format("%1$-20s %2$-20s",
							"total count", procCount[1]));
					System.out.println(new Formatter().format("%1$-20s %2$-20s",
							"progress", stats.getPercentageProgress() +"%"));
					if (stats.getEndTime()>0) {
						System.out.println(new Formatter().format("%1$-20s %2$-30s",
								"end time", new Date(stats.getEndTime())));
					}
					System.out.println(new Formatter().format("%1$-20s %2$-20s",
							"errors count", stats.getErrorsCount()));
					if (stats.getErrors()!=null && !stats.getErrors().isEmpty() &&
							cli.hasOption(CLI_RUN_PROCESS_SHORT_PARAM_ERRS)) {
//						listing errors details
						System.out.println(new Formatter().format("%1$-20s",
							"errors details"));
						for (final Entry<String, Set<ErrorEntry>> current : stats.getErrors().entrySet()) {
							System.out.println(new Formatter().format("%1$-20s",
								current.getKey()));
							if (current.getValue()!=null && !current.getValue().isEmpty()) {
								for (final ErrorEntry entry : current.getValue()) {
									System.out.println(new Formatter().format("%1$-20s %2$-20s",
											"message id", entry.getMessageId()));
									if (entry.getException()!=null) {
										entry.getException().printStackTrace(System.out);
									}
								}
							}
						}
					}
				} else {
					System.err.println("No stats available for process: " + pId);
					System.exit(1);
				}
			} catch (final ServiceException e) {
				System.err.println("Exception occurred when getting stats for process: " + pId +
						", message: " + e.getMessage());
				System.exit(1);
			}

		} catch (final ParseException e) {
			throw new InvalidParameterException(e.getMessage());
		} finally {
			if (context!=null) {
				context.close();
			}
		}
	}

}

