package pl.edu.icm.yadda.remoting.cli.service3.process;

import java.io.Serializable;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.prefs.Preferences;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.springframework.context.ConfigurableApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;

import pl.edu.icm.yadda.process.harvester.Constants;
import pl.edu.icm.yadda.remoting.cli.CliParameters;
import pl.edu.icm.yadda.remoting.cli.CommonContextHelper;
import pl.edu.icm.yadda.service2.exception.ServiceException;
import pl.edu.icm.yadda.service3.process.ProcessFacade;
import pl.edu.icm.yadda.service3.process.stats.SimplifiedErrorAwareProcessStats;
import pl.edu.icm.yadda.service3.process.stats.SimplifiedErrorEntry;
import pl.edu.icm.yadda.service3.process.stats.StatusType;

/**
 * Launcher for new processes.
 *
 * Service address and process name must be provided. Additionally multiple
 * entries in auxParam list in form key=value can be passed. Value can be
 * {@link String} or {@link String}[] in form key={str1,str2,str3}. If String is
 * in dd-MM-yyyy hh:mm:ss format, it will be recognized as {@link Date}."
 *
 *
 * @author krzys
 *
 */
public class RunProcess {

	public static final String CLI_RUN_PROCESS_SHORT_PARAM_PROC = "p";
	public static final String CLI_RUN_PROCESS_SHORT_PARAM_ERRS = "e";
	public static final String CLI_RUN_PROCESS_SHORT_PARAM_STATS_INTERV = "i";
	public static final String CLI_RUN_PROCESS_SHORT_PARAM_AUX_PARAMS = "P";

	private static final SimpleDateFormat PARAM_DATE_FORMAT = new SimpleDateFormat(
			"dd-MM-yyyy hh:mm:ss");

	private static final int checkInterval = 5000;

	/**
	 * @param args
	 */
	@SuppressWarnings("static-access")
	public static void main(final String[] args) throws Exception {
		final Options options = new Options();
		options.addOption(CliParameters.CLI_SHORT_PARAM_SERV_DESCR_LOC, true, "repository service descriptor location");
		options.addOption(CLI_RUN_PROCESS_SHORT_PARAM_PROC, true, "process to be run");
		options.addOption(CliParameters.CLI_SHORT_PARAM_HELP, false, "show help");
		options.addOption(CLI_RUN_PROCESS_SHORT_PARAM_ERRS, false, "print errors");
		options.addOption(CLI_RUN_PROCESS_SHORT_PARAM_STATS_INTERV, true, "statistics interval in ms");

		options.addOption(CliParameters.CLI_SHORT_PARAM_USER, "login", true, "user login to be used when importing to secured respository");
		options.addOption(CliParameters.CLI_SHORT_PARAM_PASSWD, "password", true, "user password to be used when importing to secured respository");
		options.addOption(CliParameters.CLI_SHORT_PARAM_DOMAIN, "domain", true, "user domain, optional");

		final Option property = OptionBuilder
				.withArgName("key=value")
				.hasArgs()
				.withDescription(
						"provide entry in the auxParams list."
								+ "Value can be String or String[] in form key={str1,str2,str3}."
								+ " If String is in "
								+ PARAM_DATE_FORMAT.toPattern()
								+ " format, it will be recognized as Date."
								+ " Multiple entries could be provided.")
				.create(CLI_RUN_PROCESS_SHORT_PARAM_AUX_PARAMS);
		options.addOption(property);

		final CommandLineParser parser = new GnuParser();

		final CommandLine cli = parser.parse(options, args);

		if (!cli.getArgList().isEmpty()) {
		    System.err.println("Unexpected parameters: " + cli.getArgList());
		    System.exit(1);
		}

		if (cli.hasOption(CliParameters.CLI_SHORT_PARAM_HELP) ||
				!cli.hasOption(CliParameters.CLI_SHORT_PARAM_SERV_DESCR_LOC) ||
				!cli.hasOption(CLI_RUN_PROCESS_SHORT_PARAM_PROC)) {
			final HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("runProcess.sh", options);
			return;
		}

		final String processName = cli.getOptionValue(CLI_RUN_PROCESS_SHORT_PARAM_PROC);
		final String[] parameters = cli.getOptionValues(CLI_RUN_PROCESS_SHORT_PARAM_AUX_PARAMS);
		final String serviceRootUrl = cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_SERV_DESCR_LOC);
		final boolean printErrors = cli.hasOption(CLI_RUN_PROCESS_SHORT_PARAM_ERRS);
		int interval = checkInterval;
		if (cli.hasOption(CLI_RUN_PROCESS_SHORT_PARAM_STATS_INTERV)) {
			try {
				interval = Integer.parseInt(cli.getOptionValue(
						CLI_RUN_PROCESS_SHORT_PARAM_STATS_INTERV));
			} catch (final NumberFormatException e) {
				System.out.println("Passed interval is not integer."
						+ " Default 5000 ms used.");
			}
		}
		Preferences.userRoot().put("serviceRootUrl", serviceRootUrl);

		final Map<String, Serializable> auxParams = new HashMap<String, Serializable>();
		if (parameters != null) {
			for (final String parameter : parameters) {
				String key;
				Serializable value;
				if (parameter.contains("=")) {
					key = parameter.substring(0, parameter.indexOf("="));
					final String auxValue = parameter.substring(
							parameter.indexOf("=") + 1).trim();
					value = auxValue;
					// try to parse as Date
					if (parseToDate(auxValue) != null) {
						value = parseToDate(auxValue);
					}
					// parse as String[]
					if (auxValue.contains("{")) {
						value = auxValue.substring(1, auxValue.length() - 1)
								.split(",");
					}

				} else {
					key = parameter;
					value = "true";
				}
				auxParams.put(key, value);
			}
		}

		if (cli.hasOption(CliParameters.CLI_SHORT_PARAM_USER)) {
    		auxParams.put(Constants.PARAM_AUTH_LOGIN, cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_USER));
    		auxParams.put(Constants.PARAM_AUTH_PASSWORD, cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_PASSWD));
    		auxParams.put(Constants.PARAM_AUTH_DOMAIN, cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_DOMAIN));
    	}

		final String  descriptorLocation= cli.getOptionValue(CliParameters.CLI_SHORT_PARAM_SERV_DESCR_LOC);

		//Preferences.userRoot().put("serviceRootUrl", serviceRootUrl);
		final String contextPath = CommonContextHelper.pickContextPath(cli);
		System.out.println("using context path: " + contextPath);

		Preferences.userRoot().put("descriptorLocation", descriptorLocation);
	        final ConfigurableApplicationContext context = new ClassPathXmlApplicationContext(contextPath);
//	    ConfigurableApplicationContext context =
//	    	new ClassPathXmlApplicationContext("classpath:pl/edu/icm/yadda/remoting/cli/cli-context.xml");
		try {
			if (!CommonContextHelper.maintainSession("process", cli, context, false)) {
	    		System.exit(1);
	    	}
			final ProcessFacade facade = (ProcessFacade) context.getBean(
					CommonContextHelper.PROCESS_MANAGER_FACADE3);
            final String pid = facade.start(processName, auxParams);

			final Timer timer = new Timer();
			timer.schedule(new TimerTask() {
				@Override
                public void run() {

					try {
						final SimplifiedErrorAwareProcessStats ps = facade
								.getSimplifiedErrorAwareProcessStats(pid);

						final StatusType status = ps.getStatus();
						if (status == StatusType.STARTING) {
							System.out.println(status);
						}

                        if (status != StatusType.STARTING) {
                            String totalCount = "?, ";
                            String percentageProgress = "";
                            if(ps.getProcessedCount()[1] > 0) { // total count > 0 so we expect full statistics
                                totalCount = ps.getProcessedCount()[1]+ ", ";
                                percentageProgress = ps.getPercentageProgress() + "%, ";
                            }

							System.out.print("\r" + status + " Progress: "
									+ ps.getProcessedCount()[0] + "/"
									+ totalCount
									+ percentageProgress
									+ "Errors: " + ps.getErrorsCount());
						}

						if (status == StatusType.FINISHED
								|| status == StatusType.INTERRUPTED) {

							final long st = ps.getStartTime();
							final long et = ps.getEndTime();
							final long duration = et - st;
							System.out.println("");
							System.out.println("Start time: " + st);
							System.out.println("End time: " + et);
							System.out.println("Duration: " + duration + " ms");
							timer.cancel();
							if (printErrors && ps.getErrorsCount() > 0) {
							    final Map<String, Set<SimplifiedErrorEntry>> errors = ps.getErrors();
							    for (final String k : errors.keySet()) {
									System.err
											.println("===============================");
									System.err.println("Element id: " + k);
									System.err
											.println("+++++++++++++++++++++++++++++++");
									System.err.println("Stack traces:");

								for (final SimplifiedErrorEntry ee : errors.get(k)) {
										System.err.println(" error = "
												+ ee.getSimplifiedException());
										System.err
												.println("+++++++++++++++++++++++++++++++");
								}
							    }

							}
						}

					} catch (final ServiceException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}
			}, 200, interval);

		} catch (final ServiceException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		} finally {
			// context.close();
		}
	}

	private static Date parseToDate(final String str) {
		try {
			// parse the parameter to Date
			return PARAM_DATE_FORMAT.parse(str);
		} catch (final ParseException pe) {
			return null;
		}
	}
}
